/*
  Author: PHCC
  Date: 28/12/2020

  TIME POWAS!

  The power of the green emerald -- inspired on the green infinity stone -- allows
  you to rewind your actions for up to 10 seconds back in time. It drains you 1 ring
  per second.

  Press BT_CUSTOM2 to rewind!

*/

local MAX_REWIND_CHAIN_LENGTH = 10 * TICRATE

// Making it local for lua optimizations
local gameGems = cw_gems
local gameChecker = cw_checker

// Solving local scoping issues
local getPlayerState
local setPlayerState
local getMobjState
local setMobjState

getPlayerState = function (player, stack)
  local state = {}
  state.powers = {}

  // Only the relevant powers, we dont want to bring back shields/invincibility/etc
  state.powers[pw_flashing] = player.powers[pw_flashing]
  state.powers[pw_carry] = player.powers[pw_carry]
  state.powers[pw_tailsfly] = player.powers[pw_tailsfly]
  state.powers[pw_spacetime] = player.powers[pw_spacetime]
  state.powers[pw_underwater] = player.powers[pw_underwater]
  state.powers[pw_nocontrol] = 1<<15

  if (player.followmobj)
    state.followmobjstate = getMobjState(player.followmobj, stack) // For Tails and Metal Sonic
  end

  state.pflags = player.pflags | PF_FULLSTASIS // stop moving

  // Interesting data to keep track of, not sure if this is necessary, though
  state.jumping = player.jumpingw
  state.secondjump = player.secondjump
  state.fly1 = player.fly1
  state.glidetime = player.glidetime
  state.climbing = player.climbing

  return state
end

setPlayerState = function (player, state, stack)
  player.powers[pw_flashing] = state.powers[pw_flashing]
  player.powers[pw_carry] = state.powers[pw_carry]
  player.powers[pw_tailsfly] = state.powers[pw_tailsfly]
  player.powers[pw_spacetime] = state.powers[pw_spacetime]
  player.powers[pw_underwater] = state.powers[pw_underwater]

  if (state.followmobjstate)
    setMobjState(player.followmobj, state.followmobjstate, stack)
  end

  player.pflags = state.pflags
  player.jumping = state.jumping
  player.secondjump = state.secondjump
  player.fly1 = state.fly1
  player.glidetime = state.glidetime
  player.climbing = state.climbing
end

// Recursively gets and stores the mobj's state
getMobjState = function (mo, stack)
  local state = {}
  if (not(stack)) // Yup, a stack. You will understand it later
    stack = {}
  end
  stack[mo] = state

  if (mo.player and not(stack[mo.player])) // avoid infinite recursion here
    state.player = getPlayerState(mo.player, stack) // Yup, we pass the stack here because we call getMobjState there too
  end

  // Alright, imma copy everything I can
  state.x = mo.x
  state.y = mo.y
  state.z = mo.z
  state.momx = mo.momx
  state.momy = mo.momy
  state.momz = mo.momz
  state.angle = mo.angle
  state.rollangle = mo.rollangle
  state.state = mo.state
  state.frame = mo.frame
  state.health = mo.health
  state.flags = mo.flags
  if (mo.type == MT_PLAYER)
    state.flags = $ | MF_NOCLIP // Lets make the player able to rewind through walls
  end
  state.flags2 = mo.flags2
  state.eflags = mo.eflags
  state.spawnpoint = mo.spawnpoint
  state.tics = mo.tics
  state.scale = mo.scale
  state.fuse = mo.fuse
  state.extravalue1 = mo.extravalue1
  state.extravalue2 = mo.extravalue2
  if (mo.tracer) // The tracer comes with us
    state.tracer = mo.tracer
    state.tracertype = mo.tracer.type
    if (stack[mo.tracer]) // This means we have already stored the state of that mobj. Avoid infinite recursion.
      state.tracerstate = stack[mo.tracer]
    else
      state.tracerstate = getMobjState(state.tracer, stack) // Why stop at the top level?
    end
  end
  if (mo.target) // Same thing for the target
    state.target = mo.target
    state.targettype = mo.target.type
    if (stack[mo.target])
      state.targetstate = stack[mo.target]
    else
      state.targetstate = getMobjState(state.target, stack)
    end
  end

  return state
end

setMobjState = function (mo, state, stack)
  if (not(stack))
    stack = {} // Stack here again
  end
  stack[mo] = state

  if (mo.player)
    setPlayerState(mo.player, state.player, stack)
  end

  P_TeleportMove(mo, state.x, state.y, state.z)
  mo.momx = state.momx
  mo.momy = state.momy
  mo.momz = state.momz
  mo.angle = state.angle
  mo.rollangle = state.rollangle
  mo.state = state.state
  mo.frame = state.frame
  mo.health = state.health
  mo.flags = state.flags
  mo.flags2 = state.flags2
  mo.eflags = state.eflags
  mo.spawnpoint = state.spawnpoint
  mo.tics = state.tics
  mo.scale = state.scale
  mo.fuse = state.fuse
  mo.extravalue1 = state.extravalue1
  mo.extravalue2 = state.extravalue2

  if (state.tracer and not(stack[state.tracer]))
    if (state.tracer.valid)
      setMobjState(state.tracer, state.tracerstate, stack)
    elseif (mo.tracer and mo.tracer.valid) // If we spawned a new tracer from a previous state, then we get it from here
      setMobjState(mo.tracer, state.tracerstate, stack)
    else // If the tracer is gone, spawn it back! Specially relevant for minecarts
      local newtracer = P_SpawnMobj(state.tracerstate.x, state.tracerstate.y, state.tracerstate.z, state.tracertype)
      setMobjState(newtracer, state.tracerstate, stack)
      mo.tracer = newtracer
    end
  else
    mo.tracer = state.tracer // Probably nil, but leave it here
  end
  if (state.target and not(stack[state.target])) // Exact same thing done with the tracer, too lazy to make it generic
    if (state.target.valid)
      setMobjState(state.target, state.targetstate, stack)
    elseif (mo.target and mo.target.valid)
      setMobjState(mo.target, state.targetstate, stack)
    else
      local newtarget = P_SpawnMobj(state.targetstate.x, state.targetstate.y, state.targetstate.z, state.targettype)
      setMobjState(newtarget, state.targetstate, stack)
      mo.target = newtarget
    end
  else
    mo.target = state.target
  end
end

local function addNewStateToPlayerChain(player, state) // If you dont understand linked lists, then what are you doing here?
  if (not(player.cw.rewindchain))
    player.cw.rewindchain = {}
  end

  local chain = player.cw.rewindchain

  if (not(chain.laststate))
    state.prevstate = chain
    chain.nextstate = state
    chain.laststate = state
  else
    state.prevstate = chain.laststate
    chain.laststate.nextstate = state
    chain.laststate = state
  end

  if (not(chain.length))
    chain.length = 1
  else
    chain.length = $ + 1
  end

  if (chain.length == MAX_REWIND_CHAIN_LENGTH) // Limit was reached, we do it FIFO and throw away the oldest state
    player.cw.rewindchain = chain.nextstate
    player.cw.rewindchain.length = chain.length - 1
    chain.nextstate.laststate = chain.laststate
    chain.nextstate.prevstate = nil // Clean references just to be sure
  end
end

local function rewind(player) // Where the magic happens
  if (player.rings - cw_gemCosts[EMERALD1].drain <= 0) return end // No rings, no back in time. But we leave you a ring so you dont die

  if (player.cw.rewindchain.length > 1 and player.cw.rewindchain.length % TICRATE == 0) // We only drain rings every second
    player.rings = $ - cw_gemCosts[EMERALD1].drain
    S_StartSound(player.mo, 162, player)
    S_StartSound(player.mo, 79, player)
  end

  local chainnode = player.cw.rewindchain.laststate

  if (not(chainnode) or not(chainnode.prevstate)) return end

  setMobjState(player.mo, chainnode) // But the real magic is here

  player.cw.rewindchain.length = $ - 1
  player.cw.rewindchain.laststate = chainnode.prevstate
  chainnode.prevstate.nextstate = nil // Clean references just to be sure 2
  chainnode.prevstate = nil // Clean references just to be sure 3
end

addHook("MobjDeath", function(target)
  if (not(gameChecker(EMERALD1))) return end

  if (target.player and target.player.cw.rewinding) // No dying while rewinding. Specially useful for minecarts.

    // But first we undo the things P_KillMobj did to us
    target.health = 1
    target.flags = $ | MF_SHOOTABLE | MF_SPECIAL
    return true
  end
end)

// This is the only one that always has to run for all players
addHook("ThinkFrame", function() // The intention is for this to work with all skins
  local gemPlayer = gameGems[EMERALD1].player
  for player in players.iterate
    if (player.mo and player.mo.valid)
      // I have to do this check here again
      if (not(player.cw))
         player.cw = {}
       end
      if (player == gemPlayer and player.cmd.buttons & BT_CUSTOM2)
        if (not(player.cw.rewinding)) // Sound fanciness
          S_StartSound(player.mo, 79, player)
        end
        player.cw.rewinding = true
        player.cw.usingGemPower = true
        rewind(player)
      else
        if (player.cw.rewinding) // Sound fanciness stop
          S_StopSound(player.mo)
        end
        player.cw.usingGemPower = false
        player.cw.rewinding = false
        player.mo.flags = $ & ~MF_NOCLIP
        player.powers[pw_nocontrol] = 0
        addNewStateToPlayerChain(player, getMobjState(player.mo))
      end
    end
  end
end)

addHook("MapChange", function(gamemap)
  for player in players.iterate
    player.cw.rewindchain = {} // Clean everything when map changes
  end
end)
