/*
  Author: PHCC
  Date: 28/12/2020

  This file is only responsible for generic handling of the game mode -- enemy counter,
  text on HUD and gem selection. All the other powers are handled in their own files
  and can even be removed and replaced at will, since they're just plugged in.

  Rules:
    - You should kill half of the enemies of the stage in order to proceeed
    - Oh, you think that's too easy? Use "killall" and then murder the entire
    Dr. Robotnik's troops!!!

  Press BT_CUSTOM1 to pick up an emerald and BT_CUSTOM2 to use its ability!
*/

local GEMS = {EMERALD1, EMERALD2, EMERALD3, EMERALD4, EMERALD5, EMERALD6, EMERALD7}

// Making it local for lua optimization
local game = cw_game
local gameGems = cw_gems
local enemiesLeft = cw_enemiesLeft
local isEnemy = cw_isEnemy
local gameChecker = cw_checker
local gameMode = cw_gameMode

// Adding a toggle for the game mode
COM_AddCommand("chaoswarmode", function(player, mode)
  if (mode == gameMode.enabled or mode == gameMode.disabled or mode == gameMode.extreme)
    if ((game.mode == gameMode.disabled and game.mapLoaded ~= nil)
     or (game.mode ~= gameMode.disabled and game.mapLoaded))
      print("There is a game currently running. Chaos war mode will be toggled on next map load.")
      game.next = mode
      return
    end

    game.mode = mode
    if (mode == gameMode.enabled)
      print("Chaos war mode enabled­. Kill half of the enemies in a stage in order to finish it.")
    elseif (mode == gameMode.extreme)
      print("Chaos war mode enabled­. Kill ALL the enemies in a stage in order to finish it.")
    else
      print("Chaos war mode disabled­")
    end
    return
  end

  print("Invalid argument. Valid options are: 0, 1, killall")
end, COM_ADMIN | COM_LOCAL)

// Adding a toggle for the game costs
COM_AddCommand("cw_costs", function(player, arg)
  if (arg == "disabled" or arg == "easy" or arg == "normal")
    if (arg == "disabled")
      cw_gemCosts = {
        [EMERALD1] = {drain = 0},
        [EMERALD2] = {drain = 0},
        [EMERALD3] = {drain = 0},
        [EMERALD4] = {drain = 0, passiveDrain = 0},
        [EMERALD5] = {drain = 0},
        [EMERALD6] = {drain = 0},
        [EMERALD7] = {drain = 0}
      }
    elseif (arg == "easy")
      cw_gemCosts = {
        [EMERALD1] = {drain = 1},
        [EMERALD2] = {drain = 7},
        [EMERALD3] = {drain = 5},
        [EMERALD4] = {drain = 1, passiveDrain = 0},
        [EMERALD5] = {drain = 5},
        [EMERALD6] = {drain = 1},
        [EMERALD7] = {drain = 0}
      }
    else
      cw_gemCosts = {
        [EMERALD1] = {drain = 1},
        [EMERALD2] = {drain = 15},
        [EMERALD3] = {drain = 10},
        [EMERALD4] = {drain = 3, passiveDrain = 1},
        [EMERALD5] = {drain = 15},
        [EMERALD6] = {drain = 2},
        [EMERALD7] = {drain = 0}
      }
    end
    return
  end

  print("Invalid argument. Valid options are: disabled, easy, normal")
end, COM_ADMIN | COM_LOCAL)

local function resetEnemyData()
  game.enemyTable = {}
  game.enemyIndex = 1
  game.enemiesOnLoad = 0
  game.enemiesKilled = 0
  game.newEnemiesKilled = 0
end

local function resetGameData()
  game.canExit = false
  game.mapLoaded = false
  game.starposttime = -1
  game.playerDidReborn = false
  resetEnemyData()
end


// Here we count all the enemies
// I'm storing them in a table cause I'll need that to kill them later
addHook("MobjSpawn", function(mo)
  if (not(gameChecker())) return end
  if (not(isEnemy(mo))) return end

  // Fixing weird enemy count behavior when map has NoReload flag
  if(game.starposttime == leveltime or (not(netgame) and game.playerDidReborn))
    resetEnemyData()
    game.starposttime = -1
    game.playerDidReborn = false
  end

  game.enemyTable[game.enemyIndex] = mo
  mo.enemyIndex = game.enemyIndex

  if (leveltime > 0)
    mo.newEnemy = true
  end

  game.enemyIndex = $ + 1

end)

// When an enemy dies, update the counter and decide if the player can finish the
// level
addHook("MobjDeath", function(target)
  if (not(gameChecker())) return end

  if (not(isEnemy(target))) return end

  game.enemyTable[target.enemyIndex] = false

  if (target.newEnemy)
    game.newEnemiesKilled = $ + 1
  else
    game.enemiesKilled = $ + 1
  end

  if (enemiesLeft() == 0 and not(game.canExit))
    cw_setExitSectorEnabled(true)
    print("You helped to bring balance to the universe and may now finish the level")
  elseif (enemiesLeft() > 0 and game.canExit) // Some new enemies popped up
    print("There are more enemies around, kill them first!")
    cw_setExitSectorEnabled(false)
  end
end)

// Reset everything when map changes. This is harmless, no need to add a check
addHook("MapChange", resetGameData)

addHook("MapLoad", function()
  if (not(gameChecker())) return end

  game.mapLoaded = true
  game.specialStage = G_IsSpecialStage()
  game.enemiesOnLoad = #game.enemyTable

  if (game.mode ~= "killall" and game.enemiesOnLoad % 2 == 1)
    game.enemiesKilled = -1 // because yes
  end

  if (game.enemiesOnLoad == 0) // There are no enemies, don't disable anything
    return
  end

  cw_setExitSectorEnabled(false)
end)

// Here we handle only the player gem selection.
// The handling of the gem's powers are done on
// each individual file that partains to that specific
// gem, since they are all directly connected to their
// player.
addHook("ThinkFrame", function()
  if (not(gameChecker())) return end

  for player in players.iterate
    if (player.mo and player.mo.valid)
        // Initializing my objects
        if (not(player.cw))
          player.cw = {}
        end
        if (not(player.mo.cw))
          player.mo.cw = {}
        end

        // Pressed the button
        if (not(player.cw.usingGemPower) and player.cmd.buttons & BT_CUSTOM1 and not(player.cw.prevcmd & BT_CUSTOM1))
          local n = 0
          if (player.cw.gem)
            n = player.cw.gem.n // Begin from where it stopped selecting
            player.cw.gem.player = nil // If player is holding a gem, remove the reference from the gem
          end

          for i = n+1,7 // Why is lua 1-indexed?
            local gem = GEMS[i]

            if (not(emeralds & gem)) continue end // You didn't collect the emerald, you can't have it

            if (player.cw.gem == gameGems[gem]) // To make sure the player doesn't select the same gem again
              player.cw.gem = nil
            elseif (not(gameGems[gem].player and gameGems[gem].player.valid and gameGems[gem].player.mo and gameGems[gem].player.mo.valid)) // Found a free one, take it
              gameGems[gem].player = player
              player.cw.gem = gameGems[gem]
              S_StartSound(player.mo, 139, player)
              break
            end
          end
        end

        // These checks will be used to fix enemy re-counting on maps with NoReload = true
        if (player.starposttime > game.starposttime)
          game.starposttime = player.starposttime
        end
        if (player.playerstate == PST_REBORN)
          game.playerDidReborn = true
        end

        // Saving the last buttons pressed
        player.cw.prevcmd = player.cmd.buttons
    end
  end
end)

local function drawCwHUD(drawfunc, player)
  if (not(game.mapLoaded) or game.specialStage) return end

  // Draw selected gem
  if (player.cw and player.cw.gem)
    local gemPatch = drawfunc.getSpritePatch("CEMG", player.cw.gem.n);
    drawfunc.drawScaled(310*FRACUNIT, 12*FRACUNIT, FRACUNIT/2, gemPatch, V_SNAPTOTOP|V_SNAPTORIGHT)
  end

  // Number of kills left
  drawfunc.drawString(258, 190, enemiesLeft(), V_SNAPTOBOTTOM|V_SNAPTORIGHT)
  drawfunc.drawString(288, 190, "left", V_SNAPTOBOTTOM|V_SNAPTORIGHT)
end

hud.add(drawCwHUD)